
/* audio_examp.q: examples for the audio, sndfile and wave modules */

import audio, sndfile, wave, ggi;

/* Config section. ***********************************************************/

/* Provide reasonable defaults for the audio input and output devices. This
   will pick Jack if it is available. You can also set the Q_AUDIO_INPUT and
   Q_AUDIO_OUTPUT environment variables, see audio.q for details. */

def AUDIO_IN = audio_in, AUDIO_OUT = audio_out;

/* Uncomment the following to set the devices explicitly. Device 0 should
   always be available. You can display the drivers and devices available on
   your system with the audio_drivers and audio_devices functions of the audio
   module. */

//def AUDIO_IN = 0, AUDIO_OUT = 0;

/* Audio sample rate, format, number of channels, buffer size and number of
   buffers for audio input and output. We assume a device capable of recording
   and playing back 16 bit integer samples at 44100 Hz in stereo, using 2048
   frame buffers. If you experience drop-outs (e.g., crackling noise in
   recorded wave, or during playback) then try to increase the buffer size. */

def RATE = 44100, FORMAT = PA_INT16, CHANNELS = 2, BUFSIZE = 2048,
  INFO = (RATE,CHANNELS,FORMAT,BUFSIZE);

/* Soundfile format. This is used when creating new soundfiles with sf_record
   or sf_save. The defaults below are for RIFF/WAV using 16 bit signed integer
   data. Change this as needed. (See sndfile.q and the libsndfile docs for a
   list of available formats.) */

def SF_FORMAT = SF_FORMAT_WAV or SF_FORMAT_PCM_16;

/* Realtime scheduling. ******************************************************/

/* Check system for available realtime thread priorities. */

private maxprio;

def POL = SCHED_FIFO, PRIO = maxprio;

testprio PRIO		= setsched this_thread 0 0 || true
			    where () = setsched this_thread POL PRIO;
			= false otherwise;

maxprio			= until (neg testprio) (+1) 1 - 1;

/* Execute a task in realtime. */

special realtime X;

realtime X		= setsched TH 0 0 || Y
			    where TH = this_thread,
			      Y = setsched TH POL PRIO || X;

/* Audio playback and recording. *********************************************/

/* To simplify matters, internally we process all wave data in the standard
   double precision (64 bit) floating point format supported by the wave
   module, regardless of the current FORMAT and SF_FORMAT settings. All data
   read from or written to the audio device or a file will be converted
   from/to the corresponding target format on the fly. */

/* Set this to true for playback looping. */

def LOOP = false;

/* Record a wave from the AUDIO_IN device and return it as a byte string. The
   given TIME parameter denotes the time in seconds to record (floating point
   values are ok). NOTE: The entire wave is stored in RAM, so you better keep
   it small. At 44100 KHz the internal 64 bit representation needs some 40 MB
   of RAM for each minute of stereo audio! */

record TIME		= realtime
			  (wave_from FORMAT (read_audio_stream IN LEN))
			    where IN:AudioStream =
			      open_audio_stream AUDIO_IN PA_READ INFO,
			      LEN = round (TIME*RATE*1.0);

/* Play back a wave on the AUDIO_OUT device. Note that the entire wave is
   written with a single call. The audio module (and PortAudio) does all the
   necessary buffering for us.  */

private play_loop;

play WAV		= realtime (play_loop OUT (wave_to FORMAT WAV))
			    where OUT:AudioStream =
			      open_audio_stream AUDIO_OUT PA_WRITE INFO;

play_loop OUT WAV	= write_audio_stream OUT WAV || play_loop OUT WAV
			    if LOOP;
			= write_audio_stream OUT WAV otherwise;

/* A scope. ******************************************************************/

/* Here we implement a simple scope which can be used to show an animated wave
   display in a GGI window while the wave is playing. */

/* Configuration of the GGI window. (WD,HT) denote the dimensions of the
   window; these values can be chosen as needed. For the purpose of displaying
   a stereo wave, the window is divided in the vertical direction. Each of the
   two channels of the wave gets a rectangle with height WH. The vertical top
   and center coordinates of the rectangle for the left and right channel are
   then given by (LT,LC) and (RT,RC), respectively. DPY denotes the target
   display, which has to be chosen in accordance with the host operating
   system (DirectX on Windows, X11 on Unix). */

def (WD,HT) = (400,240), WH = (HT-10) div 2,
  LT = 10, LC = LT+WH div 2, RT = LT+WH, RC = RT+WH div 2,
  DPY = ifelse (pos "mingw" sysinfo >= 0) "directx" "x";

/* Colors for the wave display. The defaults below are for the customary
   "radar" look. */

def BACK = (0,0,0), FORE = (0,0xffff,0);

/* The FPS value determines the number of display frames to draw per second.
   FRAMESIZE is the corresponding number of audio frames which have to be
   drawn at each iteration. */

def FPS = 10, FRAMESIZE = round (RATE/FPS);

/* Initialize the display. NOTE: Always set ASYNC mode here, to reduce
   flickering. */

init_display		= ggi_set_mode VIS (sprintf "%dx%d" (WD,HT)) ||
			  ggi_set_flags VIS GGI_FLAG_ASYNC ||
			  ggi_set_background VIS BACK ||
			  ggi_set_foreground VIS FORE ||
			  VIS
			    where VIS:GGIVisual =
			      ggi_open (sprintf "%s:-keepcursor" DPY);

/* A simple wave display function. This is useful for debugging purposes. */

display_wave WAV	= ggi_clear VIS ||
			  ggi_draw_hline VIS (0,HT div 2) WD ||
			  draw_wave VIS (0,0) (WD,HT) WAV ||
			  ggi_flush VIS || VIS
			    where VIS:GGIVisual = init_display;

/* Compute an FFT of the wave. Returns the magnitude spectrum only. This can
   be used as a transformation hook below. The XSCALE and YSCALE values let
   you choose the horizontal (frequency) and vertical (magnitude) scale of the
   FFT. */

def XSCALE = 8, YSCALE = 4;

calc_fft WAV		= prd_wave YSCALE MAG
			    where (MAG,_) = fft WAV, N = wave_size MAG,
			      MAG = sub_wave MAG 0 (round (N/XSCALE)-1);

/* Define a hook to be applied to the wave before it is displayed. This
   function is invoked with a wave and should return another one, which will
   be displayed instead. This can be used to show an FFT instead of the plain
   waveform or perform some other appropriate tranformation. Use the identity
   function to just display the waveform unchanged. */

def HOOK = id;
//def HOOK = calc_fft;

/* Display a piece of a stereo wave. */

show_wave VIS WAV	= ggi_clear VIS ||
			  ggi_draw_hline VIS (0,LC) WD ||
			  ggi_draw_hline VIS (0,RC) WD ||
			  draw_wave VIS (0,LT) (WD,WH) (HOOK L) ||
			  ggi_puts VIS (5,LT) "L" ||
			  draw_wave VIS (0,RT) (WD,WH) (HOOK R) ||
			  ggi_puts VIS (5,RT) "R" ||
			  ggi_flush VIS
			    where [L,R] = split_wave 2 WAV;

private scope_loop;

/* Here's the main function of the scope. */

scope WAV		= realtime (scope_loop VIS OUT 0 WAV)
			    where OUT:AudioStream =
			      open_audio_stream AUDIO_OUT PA_WRITE INFO,
			      VIS:GGIVisual = init_display;

/* The display/output loop. This is a fairly simplistic design. In production
   code, playback and display would be handled in separate threads (and the
   display thread wouldn't run under realtime scheduling!). See the
   qaudioplayer program on the Q website for how this is done in a "real"
   application. */

scope_loop VIS OUT I WAV
			= show_wave VIS WAV1 ||
			  write_audio_stream OUT (wave_to FORMAT WAV1) ||
			  scope_loop VIS OUT (I+2*FRAMESIZE) WAV
			    where WAV1 = sub_wave WAV I (I+2*FRAMESIZE-1)
			    if I < wave_size WAV;
			= scope_loop VIS OUT 0 WAV
			  if LOOP;
			= () otherwise;

/* Waveform generation and manipulation. *************************************/

/* This is a convenience function to create a stereo wave from a mono
   source. */

stereo_wave WAV		= join_wave [WAV,WAV];

/* Here are some simple stereo waves to play with. This also illustrates the
   use of some functions in the wave module to generate waveforms in the
   standard 64 bit floating point format. Example: play (sine 0.1 440 3) plays
   3 seconds of a sine wave at 440 Hz with an amplitude of 0.1 (10% volume). */

/*****************************************************************************
 * WARNING WARNING WARNING: These artificial sounds are *not* pleasant to    *
 * listen to. NEVER listen to them at HIGHER AMPLITUDES (always start out    *
 * with a low AMP value such as 0.1) or for EXTENDED PERIODS OF TIME, or you *
 * may DAMAGE YOUR EARS. You have been warned! ;-)                           *
 *****************************************************************************/

sine AMP FREQ TIME	= WAV
			    where // calculate number of samples
			      N = round (TIME*RATE*1.0),
			      // calculate samples per period
			      K = RATE/FREQ,
			      // generate the wave
			      WAV = prd_wave AMP (sine_wave N K),
			      // create a stereo wave (L=R)
			      WAV = stereo_wave WAV;

cosine AMP FREQ TIME	= WAV
			    where N = round (TIME*RATE*1.0),
			      K = RATE/FREQ,
			      WAV = prd_wave AMP (cosine_wave N K),
			      WAV = stereo_wave WAV;

triangle AMP FREQ TIME	= WAV
			    where N = round (TIME*RATE*1.0),
			      K = RATE/FREQ,
			      WAV = prd_wave AMP (triangle_wave N K),
			      WAV = stereo_wave WAV;

saw AMP FREQ TIME	= WAV
			    where N = round (TIME*RATE*1.0),
			      K = RATE/FREQ,
			      WAV = prd_wave AMP (saw_wave N K),
			      WAV = stereo_wave WAV;

square AMP FREQ TIME	= WAV
			    where N = round (TIME*RATE*1.0),
			      K = RATE/FREQ,
			      WAV = prd_wave AMP (square_wave N K (K/2)),
			      WAV = stereo_wave WAV;

noise AMP TIME		= WAV
			    where N = round (TIME*RATE*1.0),
			      WAV = prd_wave AMP (random_wave N),
			      WAV = stereo_wave WAV;

/* You can also create simple effects by applying basic arithmetic operations
   to a wave. For instance, here is how to create a fade in/out. Example:
   def WAV = triangle 0.2 440 2; scope (fade_in WAV++WAV++fade_out WAV) */

fade_in WAV		= WAV
			    where N = wave_size WAV,
			      WAV = prd_wave (line_wave 0 1 N) WAV;

fade_out WAV		= WAV
			    where N = wave_size WAV,
			      WAV = prd_wave (line_wave 1 0 N) WAV;

/* Sound files. **************************************************************/

/* Safe/load a wave to/from a file. */

sf_save NAME WAV	= sf_write_double SF WAV
			    where SF:SndFile = sf_open NAME SFM_WRITE
			      (RATE,CHANNELS,SF_FORMAT);

sf_load NAME		= sf_read_double SF (sf_frames SF)
			    where SF:SndFile = sf_open NAME SFM_READ ();

/* Record and play back waves to/from a file. This lets you record and play
   waves which do not fit into RAM. */

private sf_record_loop, sf_play_loop, sf_scope_loop;

sf_record NAME TIME	= realtime (sf_record_loop SF IN T0 TIME)
			    where IN:AudioStream =
			      open_audio_stream AUDIO_IN PA_READ INFO,
			      SF:SndFile = sf_open NAME SFM_WRITE
			      (RATE,CHANNELS,SF_FORMAT),
			      T0 = audio_stream_time IN;

sf_record_loop SF IN T0 T
			= sf_write_double SF
			  (wave_from FORMAT (read_audio_stream IN BUFSIZE)) ||
			  sf_record_loop SF IN T0 T
			    if audio_stream_time IN-T0 <= T;
			= () otherwise;

sf_play NAME		= realtime (sf_play_loop SF OUT)
			    where OUT:AudioStream =
			      open_audio_stream AUDIO_OUT PA_WRITE INFO,
			      SF:SndFile = sf_open NAME SFM_READ ();

sf_play_loop SF OUT	= write_audio_stream OUT (wave_to FORMAT WAV) ||
			  sf_play_loop SF OUT
			    if #WAV > 0
			    where WAV = sf_read_double SF BUFSIZE;
			= sf_seek SF 0 SEEK_SET || sf_play_loop SF OUT
			    if LOOP;
			= () otherwise;

sf_scope NAME		= realtime (sf_scope_loop VIS SF OUT)
			    where OUT:AudioStream =
			      open_audio_stream AUDIO_OUT PA_WRITE INFO,
			      SF:SndFile = sf_open NAME SFM_READ (),
			      VIS:GGIVisual = init_display;

sf_scope_loop VIS SF OUT
			= show_wave VIS WAV ||
			  write_audio_stream OUT (wave_to FORMAT WAV) ||
			  sf_scope_loop VIS SF OUT
			    if #WAV > 0
			    where WAV = sf_read_double SF FRAMESIZE;
			= sf_seek SF 0 SEEK_SET || sf_scope_loop VIS SF OUT
			    if LOOP;
			= () otherwise;
