
/* audio.q: a simple portable audio interface 05-13-03 AG
   $Id: audio.q,v 1.6 2008/01/31 07:00:16 agraef Exp $ */

/* This file is part of the Q programming system.

   The Q programming system is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   The Q programming system is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/* This module implements a (near-)realtime audio interface on top of Phil
   Burk's PortAudio library v19 (http://www.portaudio.com/). */

import clib;

/* Manifest constants. These are used for the INFO and MODE parameters of
   the open_audio_stream function. */

public var const
  // sample formats
  PA_FLOAT32, PA_INT32, PA_INT24, PA_INT16, PA_INT8, PA_UINT8,
  PA_CUSTOM_FORMAT,
  // modes
  PA_READ, PA_WRITE, PA_RDWR;

/* Default audio driver (index into the driver list). */

public var AUDIO_DRIVER;

/* Default input and output devices (indices into the device list). */

public var AUDIO_IN, AUDIO_OUT;

private extern audio_vars;

def (PA_FLOAT32, PA_INT32, PA_INT24, PA_INT16, PA_INT8, PA_UINT8,
     PA_CUSTOM_FORMAT,
     PA_READ, PA_WRITE, PA_RDWR,
     AUDIO_IN, AUDIO_OUT, AUDIO_DRIVER) = audio_vars;

/* Start and stop audio processing. Initially, audio processing is active.
   Running stop_audio closes all currently open audio streams and stops audio
   processing, start_audio restarts it again (after an automatic stop_audio if
   audio processing is currently active). This also updates the device table,
   hence this function comes in handy if new drivers and/or devices have to be
   detected at runtime. */

public extern start_audio, stop_audio;

/* Audio drivers. The audio_drivers function returns information about the
   supported driver backends (i.e., PortAudio host APIs). E.g., on Linux
   systems this might be a subset of OSS, ALSA and Jack, depending on the
   installed software and the PortAudio configuration. The result is a list of
   tuples (NAME, TYPE, DEVICES, IN, OUT), where NAME is a descriptive name of
   the driver, TYPE a unique integer identifying the driver, DEVICES the list
   of devices provided by this driver (as indices into the device list, see
   below), and IN and OUT are the default I/O devices of the driver. */

public extern audio_drivers;

/* Device info. The audio_devices function returns a list with information
   about all available audio devices of all drivers. Note that the same
   physical device may be listed more than once, if it is provided via
   different drivers. Each entry of the list is a tuple of the form (NAME,
   DRIVER, INPUT_CHANNELS, OUTPUT_CHANNELS, DEFAULT_RATE) indicating,
   respectively, the device name, the driver which provides the device (an
   index into the driver list, see above), the maximum number of input and
   output channels, and the default sample rate. */

public extern audio_devices;

/* Check whether audio processing is currently active. If this function
   returns false, most other operations of this module will fail. */

public audio_active;

audio_active			= islist audio_devices;

/* Convenience functions to determine alternate input/output devices. These
   first inspect the Q_AUDIO_IN and Q_AUDIO_OUT environment variables. If
   these are set to integers, the corresponding entries in the audio_devices
   list will be used. Alternatively, you can also specify a shell glob pattern
   matching the device name instead; thus, e.g., "SBLive*" will match the ALSA
   device for a SoundBlaster Live. If no device can be found that way, we look
   for JACK and use its default input/output ports if we can. Failing that, we
   just pick the first devices with a nonzero number of input/output ports. If
   this also fails, the PortAudio defaults (AUDIO_IN, AUDIO_OUT) are used. You
   can employ these functions to provide reasonable defaults for the input/
   output devices in your program as follows:
   def (AUDIO_IN,AUDIO_OUT) = (audio_in,audio_out); */

public audio_in, audio_out;

/* Audio stream objects. These are returned by the open_audio_stream function
   and used by the other functions. */

public extern type AudioStream;

public is_audio_stream X;

is_audio_stream _:AudioStream	= true;
is_audio_stream _		= false otherwise;

/* Open and close an audio stream. To open an audio stream you have to specify
   the index ID of the device in the device table, the mode of the stream
   (PA_READ for read-only, PA_WRITE for write-only, or PA_RDWR for read/write
   access) and the description INFO, a tuple (SAMPLE_RATE, CHANNELS, FORMAT,
   SIZE) denoting the desired sample rate, number of channels, sample format
   and buffer size, respectively.

   The SIZE parameter denotes the number of frames to be processed at a time,
   i.e., the maximum number of frames which can be read from and written to
   the stream in a single operation without blocking. This parameter might
   need some tuning in order to prevent drop-outs while achieving an
   acceptable latency. Also note that not all sample formats are supported on
   all platforms, but PA_FLOAT32, PA_INT32 and PA_INT16 should always work.

   Audio streams (and the associated devices) are closed automatically when
   the corresponding AudioStream object is garbage-collected. They can also be
   closed explicitly with the close_audio_stream function. */

public extern open_audio_stream ID MODE INFO;
public extern close_audio_stream AS;

/* Retrieve information about an audio stream. The audio_stream_id function
   returns the device index, audio_stream_info the description of the stream
   in the same format as the INFO parameter of open_audio_stream (sample rate,
   channels, etc.). The audio_stream_latencies function returns a pair
   (INPUT_LATENCY, OUTPUT_LATENCY) indicating the actual input and output
   latencies of the stream, as reported by the driver. The audio_stream_time
   function yields the current time of the stream in seconds as a floating
   point value, which is useful for synchronization purposes. The
   bytes_per_sample and bytes_per_frame functions return the number of bytes
   per sample and frame for the chosen format and number of channels, which is
   useful for decoding the data. */

public extern audio_stream_id AS;
public extern audio_stream_info AS, audio_stream_latencies AS;
public extern audio_stream_time AS;
public extern bytes_per_sample AS, bytes_per_frame AS;

/* Read and write an audio stream. Audio data is encoded as a raw byte string
   which contains a sequence of frames (interleaving samples for multichannel
   streams) in the stream's format. For the read function, NFRAMES determines
   the number of frames to be read. For the write function, only complete
   frames will be written, i.e., any extra bytes at the end of the byte string
   are ignored. Whatever the buffer size of the stream may be, you can read or
   write an arbitrary number of samples in a single operation; if the request
   cannot be satisfied immediately, the functions will block, i.e., the
   calling thread will be suspended until the given number of frames has been
   read or written. */

public extern read_audio_stream AS NFRAMES;
public extern write_audio_stream AS DATA;

/* Determine how much audio data (number of frames) could be read or written
   without blocking. */

public extern audio_stream_readable AS;
public extern audio_stream_writeable AS;

/* Implementation of audio_in/audio_out functions. */

from system import getenv;

private default_in, default_out, jack_devs, in_dev_ok, out_dev_ok, find P Xs;

audio_in
where S:String = getenv "Q_AUDIO_IN", DEVS = audio_devices:
		= ID where VAL:Int = val S, ID:Int = (DEVS!VAL)!1;
		= (DEVS!ID)!1 if ID>=0
		    where ID = find (fnmatch S.(!0)) DEVS;
otherwise:	= default_in;

audio_out
where S:String = getenv "Q_AUDIO_OUT", DEVS = audio_devices:
		= ID where VAL:Int = val S, ID:Int = (DEVS!VAL)!1;
		= (DEVS!ID)!1 if ID>=0
		    where ID = find (fnmatch S.(!0)) DEVS;
otherwise:	= default_out;

default_in	= DEVS!ID if ID>=0
		    where DEVS = jack_devs, ID = find in_dev_ok DEVS;
		= DEVS!ID if ID>=0
		    where DEVS = [0..#audio_devices-1],
		      ID = find in_dev_ok DEVS;
		= AUDIO_IN otherwise;

default_out	= DEVS!ID if ID>=0
		    where DEVS = jack_devs, ID = find out_dev_ok DEVS;
		= DEVS!ID if ID>=0
		    where DEVS = [0..#audio_devices-1],
		      ID = find out_dev_ok DEVS;
		= AUDIO_OUT otherwise;

jack_devs	= cat (map (!2) DRVS)
		    where DRVS = filter ((=12).(!1)) audio_drivers;

in_dev_ok (_,_,N,_,_)
		= N>0;
in_dev_ok ID:Int
		= in_dev_ok (audio_devices!ID);
in_dev_ok _	= false otherwise;

out_dev_ok (_,_,_,M,_)
		= M>0;
out_dev_ok ID:Int
		= out_dev_ok (audio_devices!ID);
out_dev_ok _	= false otherwise;

private find_loop N P Xs;

find P Xs	= find_loop 0 P Xs;

find_loop N P []
		= -1;
find_loop N P [X|Xs]
		= N if P X;
		= find_loop (N+1) P Xs otherwise;
