
/* wave.q: basic wave data manipulation, Sep 17 2004 AG
   $Id: wave.q,v 1.9 2005/06/25 04:41:25 agraef Exp $ */

/* This file is part of the Q programming system.

   The Q programming system is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   The Q programming system is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

import clib, ggi;

/* This module provides some basic operations for creating and manipulating
   wave data. All wave data is represented as a byte string of 64 bit floating
   point values; operations are provided to convert this format from and to
   the formats known to the audio interface. Most operations are implemented
   in C, to give the necessary performance for realtime operation. A function
   to draw wave data in a GGI visual is also provided. */

/*****************************************************************************/

/* Some utility functions. The is_wave function checks whether a byte string
   is valid wave data; this only verifies that the data is properly aligned to
   an 8 byte boundary. The wave_size function returns the size of a wave
   (number of samples). The sample function returns the sample at a given
   index in the range 0..wave_size X-1. The wave_min and wave_max functions
   determine the minimum and maximum sample value of a wave, respectively,
   while wave_rms calculates its root mean square. */

public is_wave X, wave_size X, sample X I;

is_wave X:ByteStr	= (#X mod 8=0);
is_wave _		= false otherwise;

wave_size X:ByteStr	= #X div 8 if #X mod 8=0;

sample X:ByteStr I:Int	= bfloat S where S:ByteStr = sub X (I*8) (I*8+7)
			    if (#X mod 8=0) and then (I<#X);

public extern wave_min X, wave_max X, wave_rms X;

/* Convert 64 bit wave data from/to a list of sample values. */

public extern wave Xs, samples X;

/*****************************************************************************/

/* Convert 64 bit wave data from/to one of the formats supported by the audio
   interface (PA_INT8, PA_INT16, etc.). */

public extern wave_from FORMAT X;		// FORMAT -> 64 bit
public extern wave_to FORMAT X;			// 64 bit -> FORMAT

/* Multichannel wave data is commonly represented as a single byte string,
   interleaving the values of the different channels. The following routines
   convert between this representation and a list of single-channel wave
   data. */

public extern split_wave K X, join_wave Xs;

/*****************************************************************************/

/* Basic waveform generators. A and B are integer or floating point constants,
   N is the total number of samples, and K the number of samples in one period
   (where applicable). For a square wave, L denotes the number of "on" samples
   in each period. Both K and L may also be fractional (floating point) values,
   but N must always be an integer. */

public extern cst_wave A N;			// constant-valued ("DC")
public extern line_wave A B N;			// line segment
public extern sine_wave N K;			// sine
public extern cosine_wave N K;			// cosine
public extern triangle_wave N K;		// triangular wave
public extern saw_wave N K;			// saw wave
public extern square_wave N K L;		// square wave
public extern random_wave N;			// random wave (white noise)

/* The following waveforms are commonly used for windowing a wave when
   performing frequency analysis (see also the discussion of the fft and ifft
   function below). To these ends, the wave segment to be analyzed is
   multiplied with the window. */

public extern welch_window N, bartlett_window N, hanning_window N,
  hamming_window N;

/*****************************************************************************/

/* Extract a subwave. The given offsets I and J are zero-based sample (not
   byte!) indices. */

public sub_wave X I J;

sub_wave X:ByteStr I:Int J:Int
			= Y where Y:ByteStr = sub X (8*I) (8*J+7)
			    if is_wave X;

/* Phase shifts. Shifts the phase of a wave by rotating it K samples to the
   right (K>=0) or left (K<0). */

public shift_wave K X;

shift_wave K:Int X:ByteStr
			= X if (K = 0) or else (K = N) or else (K = -N)
			    where N:Int = wave_size X;
			= sub_wave X (N-K) (N-1) ++ sub_wave X 0 (N-K-1)
			    if (K > 0) and then (K < N)
			    where N:Int = wave_size X;
			= sub_wave X (-K) (N-1) ++ sub_wave X 0 (-K-1)
			    if (K < 0) and then (K > -N)
			    where N:Int = wave_size X;

/* Reverse a wave. */

public extern reverse_wave X;

/*****************************************************************************/

/* Arithmetic. Each of the arguments may either be a constant (integer or
   floating point value) or wave data, and at least one argument *must* be
   wave data. If both arguments are wave data, they must be of the same
   size. */

public extern neg_wave X;			// negation (phase inversion)
public extern inv_wave X;			// reciprocal
public extern sum_wave X Y, diff_wave X Y;	// add/subtract
public extern prd_wave X Y, quot_wave X Y;	// multiply/divide
public extern pow_wave X Y;			// power

public extern min_wave X Y, max_wave X Y;	// min/max
public extern abs_wave X;			// absolute value
public extern sqrt_wave X;			// square root

/* Exponentials and logarithms. */

public extern exp_wave X, ln_wave X;		// base e = 2.71818...
public extern log_wave X, lg_wave X;		// base 10 and 2

/* Trigonometric and hyperbolic functions. */

public extern sin_wave X, cos_wave X, tan_wave X;
public extern asin_wave X, acos_wave X, atan_wave X;

public extern sinh_wave X, cosh_wave X, tanh_wave X;
public extern asinh_wave X, acosh_wave X, atanh_wave X;

/*****************************************************************************/

/* Sample rate conversion (SRC). These functions provide an interface to Erik
   de Castro Lopo's libsamplerate library. Requires libsamplerate 0.0.15 or
   later. See http://www.mega-nerd.com/SRC/ for more information. NOTE: As
   libsamplerate only supports single precision (32 bit), some precision will
   be lost in the conversion. */

/* Manifest constants. These constants denote the desired conversion type,
   which is passed to the src_new function. The meaning of these constants is
   as follows (these descriptions have been pilfered straight from the
   libsamplerate docs):

   SRC_SINC_BEST_QUALITY - This is a bandlimited interpolator derived from the
   mathematical sinc function and this is the highest quality sinc based
   converter, providing a worst case Signal-to-Noise Ratio (SNR) of 97
   decibels (dB) at a bandwidth of 97%. All three SRC_SINC_* converters are
   based on the techniques of Julius O. Smith although this code was developed
   independantly.

   SRC_SINC_MEDIUM_QUALITY - This is another bandlimited interpolator much like
   the previous one. It has an SNR of 97dB and a bandwidth of 90%. The speed of
   the conversion is much faster than the previous one. 

   SRC_SINC_FASTEST - This is the fastest bandlimited interpolator and has an SNR
   of 97dB and a bandwidth of 80%. 

   SRC_ZERO_ORDER_HOLD - A Zero Order Hold converter (interpolated value is
   equal to the last value). The quality is poor but the conversion speed is
   blindlingly fast.

   SRC_LINEAR - A linear converter. Again the quality is poor, but the
   conversion speed is blindingly fast. */

public const var SRC_SINC_BEST_QUALITY, SRC_SINC_MEDIUM_QUALITY,
  SRC_SINC_FASTEST, SRC_ZERO_ORDER_HOLD, SRC_LINEAR;

def SRC_SINC_BEST_QUALITY = 0,
  SRC_SINC_MEDIUM_QUALITY = 1,
  SRC_SINC_FASTEST        = 2,
  SRC_ZERO_ORDER_HOLD     = 3,
  SRC_LINEAR              = 4;

/* The SRCState type encapsulates the state information maintained during the
   conversion process. Objects of this type are created with the src_new
   function. They are then used as handles which are repeatedly passed to the
   src_process function to succesively convert all chunks of the input wave. */

public extern type SRCState;

public is_src_state X;

is_src_state X:SRCState	= true;
is_src_state _		= false otherwise;

/* Create a new SRCState object. SRC_TYPE denotes the desired conversion type
   (any of the SRC_* constants defined above) and CHANNELS the number of
   channels in the source data (each channel will be treated separately). */

public extern src_new SRC_TYPE CHANNELS;

/* Process the next chunk of wave data. Returns the converted wave. SRC is the
   SRCState object, RATIO the desired conversion ratio and X the data to be
   converted. RATIO is given as an integer or floating point value denoting
   the output rate divided by the input rate. This value may vary for
   different invokations of this function, in which case libsamplerate
   performs a linear interpolation between the old and new value (unless you
   set the new rate with the src_set_ratio function beforehand, see below). */

/* NOTE: I observed that, even if the conversion ratio and input wave size is
   constant, the actual output wave sizes may vary (sometimes even
   considerably). Apparently, this is due to rounding effects and wrapup at
   the beginning and end of the conversion process and depends on the
   installed library version and specific converter being used. Some older
   libsamplerate versions also have bugs, so make sure you use a fairly recent
   one. */

public extern src_process SRC RATIO X;

/* The src_process1 function is like src_process, but indicates that this is
   the last chunk to be processed. You can also use this when processing the
   entire wave at once. It automatically resets the source state (as with
   src_reset, see below) to prepare for the next conversion. */

public extern src_process1 SRC RATIO X;

/* Set a new conversion ratio. This allows you to bypass the linear rate
   interpolation which is usually performed by libsamplerate (see the
   description of src_process above) and obtain a step response in the
   conversion ratio. */

public extern src_set_ratio SRC RATIO;

/* Reinitialize the conversion process. Run this function to reuse an existing
   SRCState object for a new conversion process (this is also invoked
   automatically after src_process1). */

public extern src_reset SRC;

/* Simple API. The following function allows you to convert an entire wave X
   with the given conversion ratio RATIO at once. The other arguments are the
   same as for src_new. */

public src_simple SRC_TYPE CHANNELS RATIO X;

src_simple SRC_TYPE CHANNELS RATIO X
			= Y where
			      SRC:SRCState = src_new SRC_TYPE CHANNELS,
			      Y:ByteStr = src_process1 SRC RATIO X;

/*****************************************************************************/

/* Compute Fourier transforms. The fft function computes the FFT of the given
   wave data, which must not be empty. It returns a pair (MAG,PHASE) of wave
   data with the (normalized) magnitudes and phase values, respectively. Both
   MAG and PHASE consist of N/2+1 64 bit floating point values, where N is the
   size of the wave. The values at index 0 and N/2 always denote the
   components at frequency 0 ("direct current") and the Nyquist frequency
   (half the sample rate), respectively.

   The ifft function reverses the process, computing the inverse Fourier
   transform for the given (MAG,PHASE) data. This always creates an even
   number of samples. Thus, if you want to use an fft/ifft combo to manipulate
   wave data in the frequency domain then you should always work with an even
   number of samples. In this case ifft.fft is the identity (up to rounding
   errors).

   These functions are implemented using the "halfcomplex DFT" from the FFTW
   library version 3, see http://www.fftw.org/ for more details. They will
   work with an arbitrary number of samples, but in order to achieve good
   performance you should probably stick to powers of 2 if possible. */

public extern fft X, ifft Y;

/*****************************************************************************/

/* Render a wave on a GGI visual. The wave is scaled to match the given
   dimensions. The arguments are as follows:

   - VIS:	the visual to draw on;
   - P, DIM:	upper-left corner (X,Y) and dimensions (W,H) of the target
		rectangle in which to draw the wave;
   - X:		the wave data.

   The waveform is drawn in the current foreground color. At unit scale, one
   pixel is drawn for each sample. When zooming in (scale > 1), the samples
   are shown as tiny dots which are connected with straight lines. When
   zooming out (scale < 1), the range between the minimum and maximum sample
   values is shown for each pixel; moreoever, inside the waveform, the RMS
   (root mean square of sample values for each pixel) is shown in a shaded
   color. */

public extern draw_wave VIS P DIM X;
